unit game_unit;

{
    paint procedures
}

interface

uses windows,graphics,classes,extCtrls,forms;

type TGameStatus = record
                    pc1 : byte;
                    pc2 : byte;
                    pc3 : byte;
                    cnPile : byte;
                    cnLoad : byte;
                   end;

procedure createmap;
procedure freemap;
procedure initgame;
procedure showGame;
function loadStone(p : byte) : boolean;
function unloadStone(p:byte) : boolean;
procedure movestone(p1,p2 : byte);
function GetStatus : TGameStatus;

implementation

uses unit1;

const maxheight = 6;
      BGcolor = $f0ffff;
      stoneColor : array[1..maxheight] of dword =
                    ($0000ff,$00ff00,$ff0000,$0080ff,$ff00ff,$ffff00);
      minStoneWidth  = 60;
      StoneIncrement = 30;
      stoneHeight  = 30;
      pilespacing  = 40;
      pile1X       = 145;
      poleTop      = 115;
      poleWidth    = 11;
      poledistance = 250;
      railTop      = 50;
      railHeight   = 11;
      jointTop     = 20;
      jointTopheight = 16;
      craneWheelY  = jointTop + (jointTopHeight shr 1);
      craneWheelRadius  = railTop - craneWheelY;
      craneWheelBase    = 120;
      CraneleftPos  = pile1X;
      craneRightPos = pile1X+2*stoneIncrement;
      cranePosTop = 80;
      CraneWidth  = 240;
      lowerJointTop  = 65;
      lowerjointheight = 15;
      chainbase = 200;

var map : Tbitmap;
    game : array[1..3,1..maxheight] of byte;
    pheight : array[1..3] of byte;
    stonecount : byte;
//
    cranePosX : word;
    cranePosY : word;
    craneLoad : byte; //0:no load 1..6:loaded with stone ..
    cranePile : byte;
    clampBase : word;

//-- general -----------------------------------------------------------------


procedure Tickdelay(d : word);
//delay d milliseconds
//used for delays > 10 ms
var time : dword;
begin
 time := Gettickcount + d;
 while Gettickcount < time do application.processmessages;
end;

procedure CPUclockDelay(del : word);
//delay del millisecs
//used for small delays <= 10 msecs
var T1,T2 : int64;
begin
 getCPUticks(T1);
 T1 := T1 + trunc(del*CPUclock);
 repeat
  getCPUticks(T2);
  application.ProcessMessages;
 until T2 > T1;
end;

//--dimensions----------------------------------------------------------------

function GetPoleRect(p : byte) : Trect;
begin
 with result do
  begin
   left := (p-1)*poledistance + pile1x - (polewidth shr 1);
   right := left + polewidth;
   top := poleTop;
   bottom := 315;
  end;
end;

function GetStoneWidth(n : byte) : word;
//stone Nr to width
begin
 result := (n-1)*stoneIncrement+minstoneWidth;
end;

function GetStoneRect(p,n : byte) : Trect;
//calculate stone rectangle
var stoneNr , stonewidth : byte;
begin
 stoneNr := game[p,n];
 stonewidth := GetStoneWidth(stoneNr);
 with result do
  begin
   left := (p-1)*poledistance + pile1X - (stoneWidth shr 1);
   top := 315 - n*stoneheight;
   right := left + stonewidth;
   bottom := top + stoneheight;
  end;
end;

function GetUpperCraneRect : Trect;
begin
 with result do
  begin
   left := cranePosX - (craneWidth shr 1) -20;
   right := left + craneWidth + 40;
   top := 0;
   bottom := railTop;
  end;
end;

function GetLowerCraneRect : Trect;
begin
 with result do
  begin
   left := cranePosX - (craneWidth shr 1) - 20;
   right := left + cranewidth + 40;
   top := railtop + railHeight;
   bottom := cranePosY + stoneHeight;
  end;
end;

//-- level 3 paints-----------------------------------------------------------------

procedure paintRect(r : Trect; col : dword);
begin
 with map.Canvas do
  begin
   brush.Color := col;
   brush.style := bsSolid;
   fillrect(r);
  end;
end;

procedure eraseRect(r : Trect);
begin
 paintRect(r,BGcolor);
end;

procedure eraseGame;
var r : Trect;
begin
 with map do r := rect(0,0,width,height);
 paintRect(r,BGcolor);
end;

procedure paintstoneRect(r : Trect; stoneNr : byte);
//paint stone at rectangle r, color col
begin
 with map.Canvas do
  begin
   brush.Color := stonecolor[stoneNr];
   brush.style := bsSolid;
   fillrect(r);
   pen.color := $000000;
   pen.Width := 1;
   moveto(r.left,r.top);
   lineto(r.left,r.bottom-1);
   lineto(r.right-1,r.bottom-1);
   lineto(r.right-1,r.top);
   lineto(r.Left,r.top);
  end;
end;

procedure paintpoleRect(r : Trect);
var i : byte;
begin
 with map.canvas do
  begin
   pen.width := 1;
   for i := 0 to poleWidth-1 do
    begin
     pen.color := $c0c0c0 - abs(i-5)*$101010;
     moveto(r.left+i,r.top);
     lineto(r.left+i,r.bottom);
    end;
  end;
end;

procedure paintCraneRail;
var i : byte;
    right : word;
begin
 right := map.Width-10;
 with map.canvas do
  begin
   pen.Width := 1;
   for i := 0 to railheight-1 do
    begin
     pen.Color := $00a0d0 - abs(i-5)*$001414;
     moveto(10, railtop+i);
     lineto(right,railtop+i);
    end;
  end;
end;

procedure paintcranewheels;//(pos : word);
var x : word;
    x1,y1 : smallInt;
    angle : single;
    i : byte;
begin
 x := cranePosX - (craneWheelbase shr 1);
 angle := 0.045*cranePosX;
 x1 := trunc(cos(angle)*15);
 y1 := trunc(sin(angle)*15);
 with map.Canvas do
  for i := 0 to 1 do
   begin
    pen.Width := 2;
    pen.color := $000040;
    brush.Color := $0060c0;
    ellipse(x-craneWheelRadius,craneWheelY-craneWheelradius,
            x+craneWheelradius,craneWheelY+CraneWheelradius);
    brush.Color := $00ffff;
    pen.color := $c0c0c0;
    pen.Width := 1;
    ellipse(x+x1-4,craneWheelY+y1-4,x+x1+4,cranewheelY+y1+4);
    x := x + craneWheelBase;
 end;
end;

procedure paintWheelJoint;
//plus chains
var y,radius : word;
    i : byte;
    xleft,xright,Ytop,Ybottom : word;
begin
 with map.Canvas do
  begin
   pen.Width := 1;
   pen.color := $70;
   brush.Color := $40;
   xleft := cranePosX - (craneWidth shr 1);
   xRight := Xleft + craneWidth;
   yTop := jointTop;
   yBottom := jointTop + jointTopHeight;
   radius := jointTopHeight shr 1;
   ellipse(xleft-radius,yTop,xleft+radius,yBottom);
   ellipse(xright-radius,yTop,xright+radius,Ybottom);
   fillrect(rect(xleft,yTop,xRight,Ybottom));
   pen.color := $ffffff;
   xleft := cranePosX - (craneWheelBase shr 1);
   xright := xleft  + craneWheelBase;
   y := jointTop + (jointTopHeight shr 1);
   moveto(xleft-1,y-1); lineto(xleft+2,y+2);
   moveto(xleft-1,y+1); lineto(xleft+2,y-2);
   moveto(xRight-1,y-1); lineto(xRight+2,y+2);
   moveto(xRight-1,y+1); lineto(xRight+2,y-2);
//
   xleft := cranePosX - (chainBase shr 1);         //chains
   xright := xleft + chainBase;
   y := jointTop + jointTopHeight;
   for i := 0 to 8 do
    begin
     if i and 1 > 0 then pen.color := $c0c0c0 else pen.color := $000000;
     moveto(xleft+i,y); lineto(xleft+i,railtop);
     moveto(xright+i,y); lineto(xright+i,railTop);
    end;
  end;
end;

procedure paintUpperCrane;
//at cranePosX
begin
 paintCraneWheels;
 paintWheelJoint;
end;

procedure paintCraneLoad;
//craneLoad 0: exit; > 0:paint stone Nr at cranePosX,Y
var r : Trect;
    stoneWidth : byte;
begin
 stoneWidth := (craneload+1)*30;
 if craneload > 0 then
  begin
   r.Left := cranePosX- (stoneWidth shr 1);
   r.Right := r.Left + stoneWidth;
   r.Top := cranePosY;
   r.Bottom := r.Top + stoneHeight;
   paintStoneRect(r,craneLoad);
  end;
end;

procedure paintlowerCrane;
//craneposX,Y hold crane position;
//clampBase: clamps distance
//craneload=0: clamps open; craneload > 0: clamps closed
var x1,y1,x2,y2,x3,y3 : word;
    x,y,radius : word;
    i : byte;
begin
 x1 := cranePosX - (craneWidth shr 1);
 x2 := x1 + cranewidth;
 y1 := cranePosY - lowerJointHeight;
 y2 := cranePosY;
 paintrect(rect(x1,y1,x2,y2),$000060);  //joint bar
 radius := lowerJointHeight shr 1;
 with map.Canvas do
  begin
   pen.color := $000040;
   ellipse(x1-radius,y1,x1+radius,y2);  //round edges
   ellipse(x2-radius,y1,x2+radius,y2);
//                                      clamp
   x1 := cranePosX - (clampBase shr 1);
   x2 := x1 + clampbase;
   pen.color := $606060;
   brush.Color := $606060;
   if craneLoad = 0 then          //clamps out
    begin                        //left clamp
     x := x1 - 12;
     y := y1 + 1;
     x3 := x1-22;
     y3 := y1+11;
     for i := 0 to 21 do
      begin
       moveto(x3,y3+i); lineto(x,y);
       if i >= 9 then inc(y) else inc(x);
      end;
//
     x := x2 + 12;              //right clamp
     y := y1 + 1;
     x3 := x2 + 22;
     for i := 0 to 21 do
      begin
       moveto(x3,y3+i);
       lineto(x,y);
       if i >= 9 then inc(y) else dec(x);
      end;
    end;
//                                          clamps engaged
   if craneload > 0 then
    begin
     paintCraneLoad;
     pen.color := $606060;
     brush.Color := $606060;
     x := x1-1;                              //left clamp
     y := y1 +8;
     y3 := y + 25;
     for i := 0 to 14 do
      begin
       moveto(x-i,y); lineto(x-i,y3-i);
      end;

     x := x2;                                //right clamp
     for i := 0 to 14 do
      begin
       moveto(x+i,y); lineto(x+i,y3-i);
      end;
    end;

   pen.color := $ffffff;
   ellipse(x1-15,y1,x1,y2);
   ellipse(x2,y1,x2+15,y2);
//
   x1 := cranePosX - (chainbase shr 1);       //chains
   x2 := x1 + chainbase;
   y1 := railTop + railHeight;
   y2 := cranePosY - lowerJointHeight;
   for i := 0 to 8 do         //chain
    begin
     if i and 1 > 0 then pen.color := $c0c0c0 else pen.color := $000000;
     moveto(x1+i,y1); lineto(x1+i,y2);
     moveto(x2+i,y1); lineto(x2+i,y2);
    end;
  end; //with map.canvas
end;

procedure eraseUpperCrane;
var r : Trect;
begin
 r := GetUpperCraneRect;
 eraseRect(r);
end;

procedure eraseLowerCrane;
var r : Trect;
begin
 r := getLowerCraneRect;
 eraseRect(r);
end;

//--level 2 ------------------------------------------------------------------

procedure erasePile(p : byte);
var r : Trect;
begin
 r.Left := 20 + poledistance*(p-1);
 r.Right := r.Left+poledistance;
 r.top := poleTop;
 r.Bottom := 325;
 eraseRect(r);
end;

procedure paintpole(n : byte);
var r : Trect;
begin
 r := getPoleRect(n);
 paintPoleRect(r);
end;

procedure paintstone(i,j : byte);
var r : Trect;
    nr : byte;
begin
 r := getStoneRect(i,j);
 nr := game[i,j];
 paintstoneRect(r,nr);
end;

procedure paintPile(p : byte);
var i : byte;
    ch : char;
begin
 paintPole(p);
 for i := 1 to pheight[p] do paintstone(p,i);
 ch := chr(ord('A')+p-1);
 with map.Canvas do
  begin
   brush.Style := bsClear;
   font.Name := 'arial';
   font.Color := $000000;
   font.Height := 20;
   font.style := [fsBold];
   textout(140+(p-1)*poledistance,poletop,ch);
  end;
end;

procedure paintgame;
//paint complete game
var i : byte;
begin
 eraseGame;
 for i := 1 to 3 do
  begin
   paintpile(i);
  end;
 paintcranerail;
 paintuppercrane;
 paintlowercrane;
end;

procedure moveCraneHor(pos : word);
//move crane horizontally
var dv,Xinc,delay : word;
    dir : boolean;
begin
 Xinc := 5;
 delay := 10;
 dir := pos > cranePosX;
 while cranePosX <> pos do
  begin
   eraseUpperCrane;
   eraseLowerCrane;
   if dir then inc(cranePosX,Xinc) else dec(cranePosX,Xinc);
   paintUpperCrane;
   paintLowerCrane;
   paintCraneLoad;
   CPUclockdelay(delay);
   dv := abs(pos - CranePosX);      //stop
   if dv <= 20 then
    begin
     Xinc := 1;
     delay := 15;
    end;
   showGame;
  end;
end;

procedure moveCraneDown(posY,clampPos : word);
//to position of stone
var dv,Yinc,delay : word;
begin
 Yinc := 5;
 delay := 10;
 while cranePosY < posY do
  begin
   eraseLowerCrane;
   erasePile(cranePile);
   inc(cranePosY,Yinc);
   paintpile(cranePile);
   if craneLoad = 0 then
    begin
     if clampBase < clampPos then inc(clampBase,5);
     if clampBase > clampPos then dec(clampBase,5)
    end;
   paintLowerCrane;
   CPUclockdelay(delay);
   dv := abs(posY - CranePosY);
   if dv <= 20 then
    begin
     Yinc := 1;
     delay := 15;
    end;
   showGame;
  end;
end;

procedure moveCraneUp;
//to cranePosTop
var Ydec,delay : word;
begin
 Ydec := 5;
 delay := 10;
 while cranePosY > cranePosTop do
  begin
   eraseLowerCrane;
   erasePile(cranePile);
   dec(cranePosY,Ydec);
   paintpile(cranePile);
   paintLowerCrane;
   CPUclockdelay(delay);
   if cranePosY <= cranePosTop + 20 then
    begin
     Ydec := 1;
     delay := 15;
    end;
   showGame;
  end;
end;

procedure switchClamps;
begin
 tickdelay(250);
 eraseLowerCrane;
 erasepile(cranePile);
 paintPile(cranePile);
 paintlowercrane;
 ShowGame;
 tickdelay(500);
end;

//--level 1------------------------------------------------------------------

procedure createmap;
begin
 map := Tbitmap.Create;
 with map do
  begin
   pixelformat := pf32bit;
   width := form1.PaintBox1.Width;
   height := form1.PaintBox1.Height;
  end;
end;

procedure freemap;
begin
 map.Free;
end;

procedure initgame;
//init data
var i,j : byte;
begin
 stonecount := form1.UpDown1.position;
 cranePosX := craneleftPos;
 cranePosY := cranePosTop;
 craneload := 0;
 cranePile := 1;
 clampBase := minStoneWidth;
 for i := 1 to 3 do
  begin
   pheight[i] := 0;
   for j := 1 to 6 do game[i,j] := 0;
   end;
 for i := 1 to stonecount do
  begin
   inc(pheight[1]);
   game[1,pheight[1]] := stonecount - i + 1;
  end;
 paintgame;
 showGame;
end;

function GetStatus : TGameStatus;
begin
 with result do
  begin
   pc1 := pheight[1];
   pc2 := pheight[2];
   pc3 := pheight[3];
   cnpile := cranePile;
   cnLoad := craneLoad;
  end;
end;

procedure showGame;
begin
 form1.PaintBox1.Canvas.Draw(0,0,map);
end;

function LoadStone(p : byte) : boolean;
//load stone from pile p
//if pile p empty then false
var posX : word;
    r : Trect;
    stone : byte;
    stonewidth : word;
begin
 posX := pile1X + (p-1)*poledistance;
 if cranePosX <> posX then
  begin
   moveCraneHor(posX);
   tickdelay(500);
  end;
 cranePile := p;
 if pheight[p] > 0 then
  begin
   stone := game[p,pheight[p]];
   stoneWidth := GetStoneWidth(stone);
   r := getStoneRect(p,pHeight[p]);
   movecraneDown(r.Top,stoneWidth);
   showGame;
   craneLoad := stone;
   dec(pheight[p]);
   switchclamps;
   movecraneUp;
   result := true;
 end
  else result := false;//cannot load from empty pile
end;

function unloadStone(p:byte) : boolean;
//if on smaller stone then false
//move stone to top of pile p
var posX,posY : word;
begin
 posX := pile1X + (p-1)*poledistance;
 if cranePosX <> posX then
  begin
   moveCraneHor(posX);
   tickDelay(500);
  end;
 cranePile := p;
 if (pheight[p] = 0) or (game[p,pheight[p]] > craneload) then
  begin
   PosY := PoleTop+20+(5-pheight[p])*stoneHeight;
   movecraneDown(posY,0);
   inc(pheight[p]);
   game[p,pheight[p]] := craneload;
   craneload := 0;
   switchclamps;
   movecraneUp;
   result := true;
  end
   else result := false; //cannot unload on smaller stone
end;

procedure movestone(p1,p2 : byte);
//only called by recursive search
begin
 loadstone(p1);
 unloadstone(p2);
end;

end.
